#coding:utf-8

import pika
import functools
from pika.exceptions import StreamLostError, ConnectionWrongStateError


class Subscriber(object):

    def __init__(self, amqp_url, exchange, exchange_type, queue, routing_key=None, **kwargs):
        self._connection = None
        self._channel = None
        self._url = amqp_url
        self._exchange = exchange
        self._exchange_type = exchange_type
        self._routing_key = routing_key
        self._queue = queue

        # 通用配置
        self._passive = kwargs.get("passive", False)
        self._durable = kwargs.get("durable", False)
        self._auto_delete = kwargs.get("auto_delete", True)
        self._auto_ack = kwargs.get("auto_ack", True)

        # 交换机配置
        self._exchange_passive = kwargs.get("exchange_passive")
        self._exchange_durable = kwargs.get("exchange_durable")
        self._exchange_auto_delete = kwargs.get("exchange_auto_delete")
        self._exchange_auto_ack = kwargs.get("exchange_auto_ack")

        # 队列配置
        self._queue_durable = kwargs.get("queue_durable")
        self._queue_auto_delete = kwargs.get("queue_auto_delete")
        self._queue_auto_ack = kwargs.get("queue_auto_ack")

        self.connect()

    def _get_params(self, param, _type):
        prop = f"_{_type}" + param
        if hasattr(self, prop) and getattr(self, prop) is not None:
            return getattr(self, prop)
        else:
            return getattr(self, param, None)

    def connect(self):
        print('Connecting to %s', self._url)
        self._connection = pika.BlockingConnection(
            pika.URLParameters(self._url))
        self._channel = self._connection.channel()

        # 匹配交换机
        self._channel.exchange_declare(
            exchange=self._exchange,
            exchange_type=self._exchange_type,
            passive=self._get_params("_passive", "exchange"),
            durable=self._get_params("_durable", "exchange"),
            auto_delete=self._get_params("_auto_delete", "exchange")
        )

        # 声明队列
        self._channel.queue_declare(
                queue=self._queue,
                durable=self._get_params("_durable", "queue"),
                auto_delete=self._get_params("_auto_delete", "queue"))

        # 绑定队列
        self._channel.queue_bind(
            queue=self._queue, exchange=self._exchange, routing_key=self._routing_key)
        self._channel.basic_qos(prefetch_count=1)

        on_message_callback = functools.partial(
            self.on_message, userdata='on_message_userdata')

        self._channel.basic_consume(
                self._queue,
                on_message_callback,
                auto_ack=self._get_params("_auto_ack", "queue")
                )

    def reconnect(self):
        self.connect()
        print('reconnect to %s', self._url)


    def on_message(self, chan, method_frame, header_frame, body, userdata=None):
        print('Delivery properties: %s, message metadata: %s', method_frame, header_frame)
        print('Userdata: %s, message body: %s', userdata, body)
        # chan.basic_ack(delivery_tag=method_frame.delivery_tag)


    def start(self):
        while True:
            try:
                self._channel.start_consuming()
            except KeyboardInterrupt:
                self._channel.stop_consuming()
                break
            except (StreamLostError,
                    ConnectionWrongStateError,
                    ConnectionResetError):
                self.reconnect()
            except:
                raise e


    def __del__(self):
        if self._connection is not None:
            self._connection.close()


if __name__=='__main__':

    subs = Subscriber(
        'amqp://10.12.3.162:31911',
        'imlf-1',
        'direct',
        'predict',
        'predict'
    )
    subs.start()



# coding:utf-8

import os
import json
import socket
from .publish import Puber
from datetime import datetime

class LogPuber():
    def __init__(self, filepath, *args, backup=7, **kwargs):
        self.puber = Puber(*args, **kwargs)
        self.pid = os.getpid()
        self.hostname = socket.gethostname()
        self.filepath = filepath
        self.backup = backup

    def _log(self, level, *msg):
        now_str = str(datetime.now())
        if msg:
            logs = "%s|%s|host:%s|pid:%s|log:%s" % (
                    now_str,
                    level,
                    self.hostname,
                    self.pid,
                    msg[0])
            for i in msg[1:]:
                logs = "%s\n%s|%s|pid:%s|log:%s" % (
                        logs,
                        now_str,
                        level,
                        self.hostname,
                        self.pid,
                        i)
            self.puber.send(json.dumps({
                    "filepath": self.filepath,
                    "backup": self.backup,
                    "logs": logs
                }))

    def info(self, *msgs):
        self._log("INFO", *msgs)

    def error(self, *msgs):
        self._log("ERROR", *msgs)

    def warning(self, *msgs):
        self._log("WARNING", *msgs)

    def warn(self, *msgs):
        self._log("WARN", *msgs)

    def debug(self, *msgs):
        self._log("DEBUG", *msgs)


if __name__ == "__main__":
    amqp_url = 'amqp://10.12.3.162:32037'
    logger = LogPuber(
        "/root/user/logs/mq_test_one.log",
        amqp_url,
        'imlf_mqlog_test',
        'direct',
        routing_key='imlf_mqlog_key_test')
    logger.info("test log")



# coding:utf-8

import os
import json
import time
import traceback
from datetime import datetime, timedelta
from .subscribe import Subscriber

noinput_time = 600
retry_times = 3

class LogSubscriber(Subscriber):
    """订阅者
    接收日志消息
    """
    def __init__(self, *args, **kwargs):
        self.file_pool = dict()
        Subscriber.__init__(self, *args, **kwargs)

    def open_session(self, filepath, now_time):
        """打开一个文件对话, 更新调用时间"""
        if self.file_pool.get(filepath) is None:
            self.file_pool[filepath] ={
                    "f": open(filepath, "at"),
                    "utime": now_time}
        else:
            self.file_pool[filepath]["utime"] = now_time
        return self.file_pool[filepath]["f"]

    def close_session(self, filepath):
        """关闭一个文件对话"""
        if self.file_pool.get(filepath) is not None:
            try:
                file_session = self.file_pool.pop(filepath)
                file_session["f"].close()
            except:
                pass

    def reopen_session(self, filepath, now_time):
        """重新打开文件对话"""
        self.close_session(filepath)
        self.open_session(filepath, now_time)

    def close_longtime_noinput(self, now_time):
        """关闭长时间无输入的文件"""
        file_pool_filter = lambda x: now_time - self.file_pool[x]["utime"] > noinput_time
        out_of_oninput = list(filter(file_pool_filter, self.file_pool.keys()))
        for i in out_of_oninput:
            self.close_session(i)

    def __del__(self):
        for k in self.file_pool:
            self.close_session(k)

    def check_file_create_time(self, filepath, backup, now_time):
        """检查文件创建时间"""
        filectime = os.path.getctime(filepath)
        f_ctime = time.strptime(time.ctime(filectime))
        now = time.strptime(time.ctime(now_time))
        if now.tm_mday != f_ctime.tm_mday or now.tm_mon != f_ctime.tm_mon or now.tm_year != f_ctime.tm_year:
            self.close_session(filepath)
            new_filepath = '%s.%s.log' % (filepath, time.strftime("%Y-%m-%d", f_ctime))
            os.rename(filepath, new_filepath)
            self.delete_expire_file(filepath, backup)

    def delete_expire_file(self, filepath, backup):
        """删除过期的日志文件"""
        now = datetime.now()
        expire_datetime = now - timedelta(days=backup)
        while True:
            expire_datetime = expire_datetime - timedelta(days=1)
            expire_filepath = '%s.%s.log' % (filepath, expire_datetime.strftime("%Y-%m-%d"))
            if os.path.isfile(expire_filepath):
                os.remove(expire_filepath)
            else:
                break

    def write_logs(self, filepath, f, logs, now_time):
        """写日志"""
        for _ in range(retry_times):
            try:
                f.write("f:%s|%s\n" % (len(self.file_pool), logs))
                f.flush()
                break
            except:
                print(traceback.format_exc())
                self.reopen_session(filepath, now_time)

    def on_message(self, chan, method_frame, header_frame, body, userdata=None):
        """运行订阅任务"""
        try:
            # 解析数据
            json_data  = body.decode("utf-8")
            _json_data = json.loads(json_data)
            filepath   = _json_data['filepath']
            backup     = _json_data['backup']
            logs       = _json_data['logs']

            # 校验数据
            if not os.path.isabs(filepath):
                print("路径参数filepath必须是合法的绝对路径:%s" % filepath)
                return
            if not isinstance(backup, int):
                print("备份数量参数backup不是合法参数:%s" % backup)
                return
            if not logs:
                print("没有日志信息")
                return

            now_time = time.time()

            # 判断文件夹是否存在，不存在则创建
            folder = os.path.dirname(filepath)
            if not os.path.exists(folder):
                os.makedirs(folder)

            if not os.path.isfile(filepath):
                f = self.open_session(filepath, now_time)
            else:
                # 如果是新日期，则关闭旧日志文件，打开新日志文件
                self.check_file_create_time(filepath, backup, now_time)
                # 获取session
                f = self.open_session(filepath, now_time)

            # 写日志
            self.write_logs(filepath, f, logs, now_time)

            # 关闭长时间无日志的文件对话
            self.close_longtime_noinput(now_time)

        except Exception:
            print(traceback.format_exc())


if __name__=='__main__':
    amqp_url = 'amqp://10.12.3.162:32037'
    receive_log = LogSubscriber(
            amqp_url,
            'imlf_mqlog_test',
            'direct',
            'imlf_mqlog_queue_test',
            routing_key='imlf_mqlog_key_test'
            )
    receive_log.start()
